/*

                                   _JNJ`                        
                               .JNMH`                           
                              JMMF`       `;.                   
                            .NMM)          `MN.                 
                            MMM)            (MML                
                           (MMM`             MMML               
                     M  I  D  N  I  G  H  T    C  o  D  E       
                           (NMMF             MHNH               
                            NMML            .MMM                
                             NMML          .NMH                 
                              4MMNL       .#F                   
                               `4HNNL_   `                      
                                   `"""`                        

                          Copyright (C) 2004-2006
       "Ian (Larry) Latter" <ian dot latter at midnightcode dot org>

    Midnight Code is a registered trademark of Ian Latter.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, and mirrored at the Midnight Code web
    site; as at version 2 of the License only.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
    General Public License for more details.
  
    You should have received a copy of the GNU General Public License
    (version 2) along with this program; if not, write to the Free
    Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
    02111-1307 USA, or see http://midnightcode.org/gplv2.txt

*/

/* midnightcode-interface.c :: Midnight Code network interface library	*/


#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <stdlib.h>
#include <unistd.h>
#include <netdb.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <net/if.h>
#include <net/route.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "midnightcode-list.h"
#include "midnightcode-parse.h"
#include "midnightcode-config.h"
#include "midnightcode-interface.h"



int
midnightcode_interface_resolvcheck(void) {


  if(gethostbyname("www.midnightcode.org") == NULL) {
    return 1;
  }

  return 0;
}



int
midnightcode_interface_up(
                      struct midnightcode_interface_element_t * interface) {
  struct ifreq ifr;
  int sockfd;
  int one;


  if(!interface)
    return -1;

  if((sockfd = socket(AF_INET, SOCK_DGRAM, IPPROTO_IP)) < 0)
    return -1;

  one = 1;
  if(setsockopt(sockfd, SOL_SOCKET, SO_BROADCAST, &one, sizeof(one))) {
    close(sockfd);
    return -1;
  }

  memset(&ifr, 0, sizeof(struct ifreq));
  strncpy(ifr.ifr_name, interface->name, IFNAMSIZ);

  if(ioctl(sockfd, SIOCGIFFLAGS, &ifr) < 0) {
    close(sockfd);
    return -1;
  }

  /* TODO: Improve "up" check for vifs; check for IN_ADDR_ANY values */
  if(!(ifr.ifr_flags & IFF_UP)) {
    if(midnightcode_interface_configure(interface)) {
      close(sockfd);
      return -1;
    }
  } else {
    if(ioctl(sockfd, SIOCGIFFLAGS, &ifr) < 0) {
      close(sockfd);
      return -1;
    }
    interface->flags = ifr.ifr_flags;
  }

  close(sockfd);

  return 0;
}



int
midnightcode_interface_down(
                      struct midnightcode_interface_element_t * interface) {
  struct ifreq ifr;
  int sockfd;
  int one;


  if(!interface)
    return -1;

  if((sockfd = socket(AF_INET, SOCK_DGRAM, IPPROTO_IP)) < 0)
    return -1;

  one = 1;
  if(setsockopt(sockfd, SOL_SOCKET, SO_BROADCAST, &one, sizeof(one))) {
    close(sockfd);
    return -1;
  }

  memset(&ifr, 0, sizeof(struct ifreq));
  strncpy(ifr.ifr_name, interface->name, IFNAMSIZ);

  if(ioctl(sockfd, SIOCGIFFLAGS, &ifr) < 0) {
    close(sockfd);
    return -1;
  }

  if(ifr.ifr_flags & IFF_UP) {
    ifr.ifr_flags &= ~IFF_UP;
    if(ioctl(sockfd, SIOCSIFFLAGS, &ifr) < 0) {
      close(sockfd);
      return -1;
    }
  }

  if(ioctl(sockfd, SIOCGIFFLAGS, &ifr) < 0) {
    close(sockfd);
    return -1;
  }
  interface->flags = ifr.ifr_flags;

  close(sockfd);

  return 0;
}



int
midnightcode_interface_configure(
                      struct midnightcode_interface_element_t * interface) {
  struct ifreq ifr;
  struct sockaddr_in *sa;
  int sockfd;
  int one;


  if(!interface)
    return -1;

  if((sockfd = socket(AF_INET, SOCK_DGRAM, IPPROTO_IP)) < 0)
    return -1;

  one = 1;
  if(setsockopt(sockfd, SOL_SOCKET, SO_BROADCAST, &one, sizeof(one))) {
    close(sockfd);
    return -1;
  }

  memset(&ifr, 0, sizeof(struct ifreq));
  strncpy(ifr.ifr_name, interface->name, IFNAMSIZ);

  sa = (struct sockaddr_in *)&ifr.ifr_addr;
  sa->sin_family = AF_INET;
  sa->sin_port = 0;
  sa->sin_addr = interface->ipv4_address;

  if(ioctl(sockfd, SIOCSIFADDR, &ifr)) {
    close(sockfd);
    return -1;
  }

  if(interface->ipv4_mask.s_addr) {
    sa->sin_addr = interface->ipv4_mask;
    if(ioctl(sockfd, SIOCSIFNETMASK, &ifr)) {
      close(sockfd);
      return -1;
    }

    sa->sin_addr = interface->ipv4_address;
    if(ioctl(sockfd, SIOCSIFADDR, &ifr)) {
      close(sockfd);
      return -1;
    }
  }

  if(interface->ipv4_broadcast.s_addr) {
    sa->sin_addr = interface->ipv4_broadcast;
    if(ioctl(sockfd, SIOCSIFBRDADDR, &ifr)) {
      close(sockfd);
      return -1;
    }
  }

  if(!interface->is_alias) {
    if(interface->mtu) {
      ifr.ifr_mtu = interface->mtu;
      if(ioctl(sockfd, SIOCGIFMTU, &ifr)) {
        close(sockfd);
        return -1;
      }
    }
  }

  ifr.ifr_flags = interface->flags | IFF_UP | IFF_RUNNING | IFF_BROADCAST;
  if(ioctl(sockfd, SIOCSIFFLAGS, &ifr)) {
    close(sockfd);
    return -1;
  }
  interface->flags = ifr.ifr_flags;

  close(sockfd);

  return 0;
}



int
midnightcode_interface_list(struct midnightcode_interface_head_t * elements) {
  struct ifreq *ifr;
  struct ifconf ifc;
  struct sockaddr_in sa;
  u_char * uchar_p;
  int sockfd;
  int size;
  int ifrsize;
  int count;
  struct midnightcode_interface_element_t * now_element;


  size = 1;
  ifrsize = ((int)(size * sizeof (struct ifreq)));

  if((sockfd = socket(AF_INET, SOCK_DGRAM, IPPROTO_IP)) < 0) {
    return -1;
  }

  ifc.ifc_len = ifrsize;
  ifc.ifc_req = NULL;
  count = 0;

  do {
    ++size;
    ifrsize = ((int)(size * sizeof (struct ifreq)));
    if ((ifc.ifc_req = realloc(ifc.ifc_req, ifrsize)) == NULL) {
      return -1;
    }
    ifc.ifc_len = ifrsize;
    if(ioctl(sockfd, SIOCGIFCONF, &ifc)) {
      return -1;
    }
  } while(ifrsize <= ifc.ifc_len);

  ifr = ifc.ifc_req;
  for(; (char *)ifr < (char *)ifc.ifc_req + ifc.ifc_len; ifr++) {
    if(ifr->ifr_addr.sa_data == (ifr+1)->ifr_addr.sa_data)
      continue;

    now_element = malloc(sizeof(struct midnightcode_interface_element_t));
    memset(now_element, 0, sizeof(struct midnightcode_interface_element_t));

    if(ioctl(sockfd, SIOCGIFINDEX, ifr)) {
      free(now_element);
      continue;
    }
    now_element->ifindex = ifr->ifr_ifindex;

    if(!ioctl(sockfd, SIOCGIFHWADDR, ifr)) {
      uchar_p = (u_char *)&ifr->ifr_hwaddr.sa_data[0];
      snprintf(now_element->mac_address_s, 31, "%02X:%02X:%02X:%02X:%02X:%02X",
            *uchar_p, *(uchar_p + 1), *(uchar_p + 2),
            *(uchar_p + 3), *(uchar_p + 4), *(uchar_p + 5));
    }
    memcpy(&now_element->mac_address, &ifr->ifr_hwaddr.sa_data[0], IFHWADDRLEN);

    if(ioctl(sockfd, SIOCGIFADDR, ifr)) {
      free(now_element);
      continue;
    }
    strncpy(now_element->name, ifr->ifr_name, sizeof(now_element->name));
    strncpy(now_element->ipv4_address_s,
      inet_ntoa(
        (*(struct in_addr *)&ifr->ifr_addr.sa_data[sizeof sa.sin_port])
      ), 31);
    now_element->ipv4_address = 
      (*(struct in_addr *)&ifr->ifr_addr.sa_data[sizeof sa.sin_port]);

    if(ioctl(sockfd, SIOCGIFNETMASK, ifr)) {
      free(now_element);
      continue;
    }
    strncpy(now_element->ipv4_mask_s,
      inet_ntoa(
        (*(struct in_addr *)&ifr->ifr_netmask.sa_data[sizeof sa.sin_port])
      ), 31);
    now_element->ipv4_mask = 
      (*(struct in_addr *)&ifr->ifr_netmask.sa_data[sizeof sa.sin_port]);

    if(ioctl(sockfd, SIOCGIFBRDADDR, ifr)) {
      free(now_element);
      continue;
    }
    strncpy(now_element->ipv4_broadcast_s,
      inet_ntoa(
        (*(struct in_addr *)&ifr->ifr_broadaddr.sa_data[sizeof sa.sin_port])
      ), 31);
    now_element->ipv4_broadcast = 
      (*(struct in_addr *)&ifr->ifr_broadaddr.sa_data[sizeof sa.sin_port]);

    if(ioctl(sockfd, SIOCGIFFLAGS, ifr)) {
      free(now_element);
      continue;
    }
    now_element->flags = (int)ifr->ifr_flags;

    if(ioctl(sockfd, SIOCGIFMTU, ifr)) {
      free(now_element);
      continue;
    }
    now_element->mtu = ifr->ifr_mtu;

    /* TODO: true alias detection - :: is not alias */
    if(midnightcode_parse_any(now_element->name, ":"))
      now_element->is_alias = 1;

    midnightcode_list_append(elements, now_element);

    count++;
  }
  close(sockfd);

  return count;
}



int
midnightcode_interface_byindex(struct midnightcode_interface_head_t * elements,
                               int ifindex) {
  struct midnightcode_interface_element_t * now_element;
  struct midnightcode_interface_element_t * old_element;


  if(midnightcode_interface_list(elements) < 0)
    return -1;

  old_element = NULL;
  midnightcode_list_for_each(now_element, elements) {
    if(old_element != NULL) {
      free(old_element);
      old_element = NULL;
    }
    if(now_element->ifindex != ifindex || now_element->is_alias == 1) {
      midnightcode_list_remove(elements, now_element);
      old_element = now_element;
    }
  }
  if(old_element != NULL)
    free(old_element);

  return 0;
}



int
midnightcode_interface_byname(
                          struct midnightcode_interface_head_t * elements,
                          char * pattern,
                          struct midnightcode_interface_element_t ** element) {
  struct midnightcode_interface_element_t * found_element;


  if(!elements || !pattern)
    return -1;

  if(strlen(pattern) < 1)
    return -1;

  midnightcode_list_for_each(found_element, elements) {
    if(strcmp(found_element->name, pattern) == 0) {
      *element = found_element;
      return 1;
    }
  }

  return 0;
}



int
midnightcode_interface_only(struct midnightcode_interface_head_t * elements) {
  struct midnightcode_interface_element_t * now_element;
  struct midnightcode_interface_element_t * old_element;


  old_element = NULL;
  midnightcode_list_for_each(now_element, elements) {
    if(old_element != NULL) {
      free(old_element);
      old_element = NULL;
    }
    if(strncmp(now_element->name, "eth", 2) != 0) {
      midnightcode_list_remove(elements, now_element);
      old_element = now_element;
    }
  }
  if(old_element != NULL)
    free(old_element);

  return 0;
}



int
midnightcode_interface_to_config(
                     struct midnightcode_interface_head_t * interface_elements,
                     struct midnightcode_config_head_t * config_elements) {
  struct midnightcode_interface_element_t * interface_element;


  if(!interface_elements || !config_elements)
    return -1;

  if(midnightcode_config_del_structure(config_elements,
                                       MIDNIGHTCODE_INTERFACE_STRUCTURE))
    return -1;

  midnightcode_list_for_each(interface_element, interface_elements) {
    if(midnightcode_config_add_pair(config_elements,
                                    MIDNIGHTCODE_INTERFACE_STRUCTURE,
                                    interface_element->name,
                                    "is_enabled",
                                    &interface_element->is_enabled,
                                    MIDNIGHTCODE_PARSE_TYPE_INT))
      return -1;

    if(midnightcode_config_add_pair(config_elements,
                                    MIDNIGHTCODE_INTERFACE_STRUCTURE,
                                    interface_element->name,
                                    "is_alias",
                                    &interface_element->is_alias,
                                    MIDNIGHTCODE_PARSE_TYPE_INT))
      return -1;

    if(midnightcode_config_add_pair(config_elements,
                                    MIDNIGHTCODE_INTERFACE_STRUCTURE,
                                    interface_element->name,
                                    "flags",
                                    &interface_element->flags,
                                    MIDNIGHTCODE_PARSE_TYPE_INT))
      return -1;

    if(midnightcode_config_add_pair(config_elements,
                                    MIDNIGHTCODE_INTERFACE_STRUCTURE,
                                    interface_element->name,
                                    "ifindex",
                                    &interface_element->ifindex,
                                    MIDNIGHTCODE_PARSE_TYPE_INT))
      return -1;

    if(midnightcode_config_add_pair(config_elements,
                                    MIDNIGHTCODE_INTERFACE_STRUCTURE,
                                    interface_element->name,
                                    "mac_address",
                                    &interface_element->mac_address,
                                    MIDNIGHTCODE_PARSE_TYPE_HWADDR))
      return -1;

    if(midnightcode_config_add_pair(config_elements,
                                    MIDNIGHTCODE_INTERFACE_STRUCTURE,
                                    interface_element->name,
                                    "is_static",
                                    &interface_element->is_static,
                                    MIDNIGHTCODE_PARSE_TYPE_INT))
      return -1;

    if(midnightcode_config_add_pair(config_elements,
                                    MIDNIGHTCODE_INTERFACE_STRUCTURE,
                                    interface_element->name,
                                    "ipv4_address",
                                    interface_element->ipv4_address_s,
                                    MIDNIGHTCODE_PARSE_TYPE_STRING))
      return -1;

    if(midnightcode_config_add_pair(config_elements,
                                    MIDNIGHTCODE_INTERFACE_STRUCTURE,
                                    interface_element->name,
                                    "ipv4_mask",
                                    interface_element->ipv4_mask_s,
                                    MIDNIGHTCODE_PARSE_TYPE_STRING))
      return -1;

    if(midnightcode_config_add_pair(config_elements,
                                    MIDNIGHTCODE_INTERFACE_STRUCTURE,
                                    interface_element->name,
                                    "ipv4_broadcast",
                                    interface_element->ipv4_broadcast_s,
                                    MIDNIGHTCODE_PARSE_TYPE_STRING))
      return -1;

    if(midnightcode_config_add_pair(config_elements,
                                    MIDNIGHTCODE_INTERFACE_STRUCTURE,
                                    interface_element->name,
                                    "mtu",
                                    &interface_element->mtu,
                                    MIDNIGHTCODE_PARSE_TYPE_INT))
      return -1;

  }

  return 0;
}



int
midnightcode_interface_from_config(
                     struct midnightcode_interface_head_t * interface_elements,
                     struct midnightcode_config_head_t * config_elements) {
  struct midnightcode_interface_element_t * interface_element;
  struct midnightcode_config_section_element_t * section_element;
  struct midnightcode_config_section_element_t * old_element;
  struct midnightcode_config_section_head_t * section_elements;
  void * tmp_void_p;
  u_char * uchar_p;
  midnightcode_list_init(config_section_struct_head,
                         midnightcode_config_section_element_t);


  if(!interface_elements || !config_elements)
    return -1;

  tmp_void_p = (void *)&config_section_struct_head;
  section_elements = tmp_void_p;

  /* TODO: free section_elements memory */

  if(!midnightcode_config_get_sections(config_elements,
                                       section_elements,
                                       MIDNIGHTCODE_INTERFACE_STRUCTURE))
    return -1;

  old_element = NULL;
  midnightcode_list_for_each(section_element, section_elements) {
    if(old_element != NULL) {
      free(old_element);
      old_element = NULL;
    }
    old_element = section_element;

    interface_element = malloc(
                          sizeof(struct midnightcode_interface_element_t));
    memset(interface_element, 0,
      sizeof(struct midnightcode_interface_element_t));

    snprintf(interface_element->name, 63, "%s", section_element->section);

    if(midnightcode_config_get_pair(config_elements,
                                    MIDNIGHTCODE_INTERFACE_STRUCTURE,
                                    interface_element->name,
                                    "is_enabled",
                                    &interface_element->is_enabled, 0,
                                    MIDNIGHTCODE_PARSE_TYPE_INT))
      return -1;

    if(midnightcode_config_get_pair(config_elements,
                                    MIDNIGHTCODE_INTERFACE_STRUCTURE,
                                    interface_element->name,
                                    "is_alias",
                                    &interface_element->is_alias, 0,
                                    MIDNIGHTCODE_PARSE_TYPE_INT))
      return -1;

    if(midnightcode_config_get_pair(config_elements,
                                    MIDNIGHTCODE_INTERFACE_STRUCTURE,
                                    interface_element->name,
                                    "flags",
                                    &interface_element->flags, 0,
                                    MIDNIGHTCODE_PARSE_TYPE_INT))
      return -1;

    if(midnightcode_config_get_pair(config_elements,
                                    MIDNIGHTCODE_INTERFACE_STRUCTURE,
                                    interface_element->name,
                                    "ifindex",
                                    &interface_element->ifindex, 0,
                                    MIDNIGHTCODE_PARSE_TYPE_INT))
      return -1;

    if(midnightcode_config_get_pair(config_elements,
                                    MIDNIGHTCODE_INTERFACE_STRUCTURE,
                                    interface_element->name,
                                    "mac_address",
                                    &interface_element->mac_address, 0,
                                    MIDNIGHTCODE_PARSE_TYPE_HWADDR))
      return -1;

    uchar_p = interface_element->mac_address;
    snprintf(interface_element->mac_address_s, 31,
            "%02X:%02X:%02X:%02X:%02X:%02X",
            *uchar_p, *(uchar_p + 1), *(uchar_p + 2),
            *(uchar_p + 3), *(uchar_p + 4), *(uchar_p + 5));

    if(midnightcode_config_get_pair(config_elements,
                                    MIDNIGHTCODE_INTERFACE_STRUCTURE,
                                    interface_element->name,
                                    "is_static",
                                    &interface_element->is_static, 0,
                                    MIDNIGHTCODE_PARSE_TYPE_INT))
      return -1;

    if(midnightcode_config_get_pair(config_elements,
                                    MIDNIGHTCODE_INTERFACE_STRUCTURE,
                                    interface_element->name,
                                    "ipv4_address",
                                    &interface_element->ipv4_address, 0,
                                    MIDNIGHTCODE_PARSE_TYPE_IP))
      return -1;

    inet_ntop(AF_INET, &interface_element->ipv4_address,
      interface_element->ipv4_address_s, 31);

    if(midnightcode_config_get_pair(config_elements,
                                    MIDNIGHTCODE_INTERFACE_STRUCTURE,
                                    interface_element->name,
                                    "ipv4_mask",
                                    &interface_element->ipv4_mask, 0,
                                    MIDNIGHTCODE_PARSE_TYPE_IP))
      return -1;

    inet_ntop(AF_INET, &interface_element->ipv4_mask,
      interface_element->ipv4_mask_s, 31);

    if(midnightcode_config_get_pair(config_elements,
                                    MIDNIGHTCODE_INTERFACE_STRUCTURE,
                                    interface_element->name,
                                    "ipv4_broadcast",
                                    &interface_element->ipv4_broadcast, 0,
                                    MIDNIGHTCODE_PARSE_TYPE_IP))
      return -1;

    inet_ntop(AF_INET, &interface_element->ipv4_broadcast,
      interface_element->ipv4_broadcast_s, 31);

    if(midnightcode_config_get_pair(config_elements,
                                    MIDNIGHTCODE_INTERFACE_STRUCTURE,
                                    interface_element->name,
                                    "mtu",
                                    &interface_element->mtu, 0,
                                    MIDNIGHTCODE_PARSE_TYPE_INT))
      return -1;


    midnightcode_list_append(interface_elements, interface_element);

  }
  if(old_element != NULL)
    free(old_element);

  return 0;
}



/*
int
midnightcode_interface_serialize(
                          struct midnightcode_interface_head_t * elements,
                          char * buffer) {
  struct midnightcode_interface_element_t * now_element;
  struct midnightcode_interface_element_t * old_element;


  old_element = NULL;
  midnightcode_list_for_each(now_element, elements) {
    if(old_element != NULL) {
      free(old_element);
      old_element = NULL;
    }
    if(strncmp(now_element->name, "eth", 2) != 0) {
      midnightcode_list_remove(elements, now_element);
      old_element = now_element;
    }
  }
  if(old_element != NULL)
    free(old_element);

  return 0;
}
*/

